#include "arch.h"
#include "exec.h"
#include "extr.h"
#include "fs.h"
#include "link.h"
#include "main.h"
#include "math.h"
#include "mem.h"
#include "msg.h"
#include "path.h"
#include "str.h"

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>


const unsigned char Tries = 5;  /* values 1 - 9 are allowed */


static void cleanup(const struct archive *arch)
	{
	if(!arch->links.file.p)
		free(arch->links.file.e);
	}

static bool setup(struct archive *arch)
	{
	size_t len, ext, dot = 0, suf = 0;

	switch(file_type(true, arch->file.p))
		{
		case 0: err(MSG_NOT_EXIST, false, arch->file.p);
		return false;
		case 1: err(MSG_IS_DIR, false, arch->file.p);
		case -1: return false;
		}

	arch->file.e = ent_from_path(arch->file.p);
	arch->file.z = arch->file.e + strlen(arch->file.e);

	len = elen(arch->file);
	if(Extension)
		{
		ext = strlen(Extension);
		if(len < ext + 2 ||
		   arch->file.z[-ext-1] != '.' ||
		   strcasecmp(arch->file.z - ext, Extension) != 0)
			len += (dot = 1) + (suf = ext);
		}
	arch->links.file.e = alloc(len + 1);
	arch->links.file.z = str_cp_max(suf, str_cp_max(dot, str_cp(arch->links.file.e, arch->file.e), "."), Extension);

	arch->extr = find_extr(arch);
	if(!arch->extr)
		{
		err(MSG_MATCH_EXTR, false, arch->file.p);
		cleanup(arch);
		return false;
		}

	arch->pos = arch->extr->pos;
	str_tr(tolower, arch->links.file.e + arch->pos);
	return true;
	}

static char *prefix_num(unsigned int try, char *buf, const char *ent, size_t len)
	{
	switch(try)
		{
		case 1: buf = str_cp(buf, "1-");
		case 0: return str_cp_max(len, buf, ent);
		}
	*buf = '0' + try;
	return buf + 2 + len;
	}

static void cleanup_sandbox(const struct archive *arch)
	{
	free(arch->sandbox.p);
	}

static bool create_sandbox(struct archive *arch)
	{
	unsigned int i;
	size_t len;

	len = Absolute ? dlen(arch->file) : 0;
	arch->sandbox.p = alloc(len + 2 + arch->pos);
	arch->sandbox.e = str_cp_max(len, arch->sandbox.p, arch->file.p);

	for(i = 0; i < Tries; ++i)
		{
		arch->sandbox.z = prefix_num(i, arch->sandbox.e, arch->links.file.e, arch->pos - 1);
		if(mkdir(arch->sandbox.p, S_IRWXU | S_IRWXG | S_IRWXO) == 0)
			return true;
		if(errno != EEXIST)
			break;
		}
	err(MSG_MKDIR, true, arch->sandbox.p);
	cleanup_sandbox(arch);
	return false;
	}

static void exec_extr(void *args)
	{
	struct archive *arch = (struct archive *)args;
	char *name, *cmd[] = { arch->extr->extract, arch->links.file.e, NULL };

	if(chdir(arch->sandbox.p) != 0)
		die(MSG_CHDIR, true, arch->sandbox.p);
	name = alloc(5 + arch->pos);
	str_cp_max(arch->pos - 1, str_cp(name, "Name="), arch->file.e);
	if(!set_env(name))
		exit(1);
	if(cmd[1][0] == '-')
		*(cmd[1] -= 2) = '.';
	exec((void **)cmd);
	}

static bool norm_perms(const struct archive *arch)
	{
	const char *args[] = { "chmod", "-R", "a=,+rwX", arch->sandbox.p, NULL };

	return arch->extr->perms ?
	       spawn(exec, (void *)args, MSG_NORM_PERMS, arch->sandbox.p) :
	       true;
	}

static int inc_rename(unsigned int tries, const char *src, char *dst,
                      size_t offset, const char *ent, size_t len)
	{
	unsigned int i;

	for(i = 0; i < tries; ++i)
		{
		prefix_num(i, dst + offset, ent, len);
		switch(safe_rename(src, dst))
			{
			case -1: return 0;
			case  1: return i + 1;
			}
		}
	return 0;
	}

static char *move_lower(const struct path dir, const char *ent)
	{
	/*
	 * rename parent/dir/ent to parent/ent (or parent/n-ent)
	 * rmdir parent/dir
	 * try to remove n- prefix (or lower n)
	 */

	char *p, *q;
	size_t len;
	unsigned int try;

	len = strlen(ent);
	p = alloc(dlen(dir) + elen(dir) + 1 + len + 1);  /* elen(dir) + 1 >= 2 */
	str_cp(str_cp(str_cp(p, dir.p), "/"), ent);
	q = alloc(dlen(dir) + 2 + len + 1);
	str_cp_max(dlen(dir), q, dir.p);

	try = inc_rename(Tries + 1, p, q, dlen(dir), ent, len);
	if(!try)
		{
		err(MSG_RENAME, true, p);
		free(p);
		free(q);
		return NULL;
		}
	if(rmdir(dir.p) != 0)
		err(MSG_DEL, true, dir.p);
	try = inc_rename(MIN(try - 1, Tries), q, p, dlen(dir), ent, len);
	free(try ? q : p);
	return try ? p : q;
	}

static bool handle_contents(const struct archive *arch)
	{
	DIR *dir;
	struct dirent *ent;
	unsigned int entries = 0;
	char *first, *moved;

	dir = opendir_verb(arch->sandbox.p);
	if(!dir)
		return false;
	while((ent = readdir(dir)))
		{
		if(is_dot(ent->d_name))
			continue;
		if(++entries > 1 || arch->extr->subdir)
			break;
		first = strdup(ent->d_name);
		}
	closedir(dir);

	if(entries == 0)
		{
		err(MSG_EMPTY, false, arch->file.p);
		return rmdir(arch->sandbox.p) == 0 || err(MSG_DEL, true, arch->sandbox.p);
		}
	if(entries == 1 && !arch->extr->subdir)
		{
		moved = move_lower(arch->sandbox, first);
		free(first);
		if(moved)
			{
			pretty_print(moved);
			free(moved);
			}
		return moved;
		}
	/* entries > 1 || arch->extr->subdir */
	free(first);
	pretty_print(arch->sandbox.p);
	return true;
	}

static bool del_archive(const struct archive *arch)
	{
	bool success = true;
	char *path, *ent;
	const struct link *l;
	size_t longest = 0;

	if(!Unlink)
		return true;

	for(l = &arch->links; l; l = l->next)
		longest = MAX(longest, elen(l->target));
	path = alloc(dlen(arch->file) + longest + 1);
	ent = str_cp_max(dlen(arch->file), path, arch->file.p);

	for(l = &arch->links; l; l = l->next)
		{
		str_cp(ent, l->target.e);
		if(unlink(path) != 0)
			{
			err(MSG_DEL, true, path);
			success = false;
			}
		}

	free(path);
	return success;
	}

bool extract(const char *arg)
	{
	bool success = false, extr;
	struct archive arch = { (char *)arg };

	if(setup(&arch))
		{
		if(create_sandbox(&arch))
			{
			if(create_links(&arch))
				{
				extr = spawn(exec_extr, (void *)&arch, MSG_EXTRACT, arch.file.p);
				if((extr || Keep) &&
				   del_links(&arch) &&
				   norm_perms(&arch) &&
				   handle_contents(&arch))
					success = extr && del_archive(&arch);
				else
					rm_rf(arch.sandbox.p);
				cleanup_links(&arch);
				}
			cleanup_sandbox(&arch);
			}
		cleanup(&arch);
		}
	return success;
	}
