#include "arch.h"
#include "extr.h"
#include "fs.h"
#include "link.h"
#include "main.h"
#include "mem.h"
#include "msg.h"
#include "path.h"
#include "str.h"

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <unistd.h>


void cleanup_links(const struct archive *arch)
	{
	struct link *l, *next;

	free(arch->links.file.p);
	free(arch->links.target.p);

	for(l = arch->links.next; l; l = next)
		{
		free(l->file.p);
		free(l->target.p);
		next = l->next;
		free(l);
		}
	}

static void complete_first_link(struct archive *arch)
	{
	char *old = arch->links.file.e;
	size_t dot, dir;

	/* we are finally in possession of the sandbox name needed to build the link path */
	arch->links.file.p = alloc(plen(arch->sandbox) + 1 + elen(arch->links.file) + 1);
	arch->links.file.e = str_cp(str_cp(arch->links.file.p, arch->sandbox.p), "/");
	arch->links.file.z = str_cp(arch->links.file.e, old);
	free(old);

	/* build the archive link target */
	dot = Absolute || arch->file.p[0] != '/' ? 3 : 0;
	dir = Absolute ? 0 : dlen(arch->file);
	arch->links.target.p = alloc(dot + dir + elen(arch->file) + 1);
	arch->links.target.e = str_cp_max(dir, str_cp_max(dot, arch->links.target.p, "../"), arch->file.p);
	arch->links.target.z = str_cp(arch->links.target.e, arch->file.e);
	}

static bool is_aux(const struct archive *arch, const char *name)
	{
	if(strncmp(arch->links.file.e, name, arch->pos) != 0 ||
	   strcasecmp(arch->links.file.e + arch->pos, name + arch->pos) == 0)
		return false;  /* main name differs or same extension */
	match_extr(arch->extr, name);
	return arch->extr->pos == arch->pos;
	}

static void add_link(struct archive *arch, const char *name)
	{
	struct link *l;
	size_t len;

	len = strlen(name) + 1;

	l = alloc(sizeof(struct link));
	l->next = arch->links.next;
	arch->links.next = l;

	l->file.p = alloc(dlen(arch->links.file) + len);
	l->file.e = str_cp_max(dlen(arch->links.file), l->file.p, arch->links.file.p);
	l->file.z = str_cp_tr(tolower, str_cp_max(arch->extr->pos, l->file.e, name), name + arch->extr->pos);

	l->target.p = alloc(dlen(arch->links.target) + len);
	l->target.e = str_cp_max(dlen(arch->links.target), l->target.p, arch->links.target.p);
	l->target.z = str_cp(l->target.e, name);
	}

static bool make_symlink(struct link *l)
	{
	if(symlink(l->target.p, l->file.p) != 0)
		err(MSG_SYMLINK, true, l->file.p);
	else if(!get_inode(&l->inode, l->file.p))
		err(MSG_STAT, true, l->file.p);
	else
		return true;
	return false;
	}

bool create_links(struct archive *arch)
	{
	struct link *l;
	DIR *dir;
	struct dirent *ent;

	dir = open_path_dir(arch->file);
	if(!dir)
		return false;
	complete_first_link(arch);
	while((ent = readdir(dir)))
		if(is_aux(arch, ent->d_name))
			add_link(arch, ent->d_name);
	closedir(dir);

	for(l = &arch->links; l; l = l->next)
		if(!make_symlink(l))
			{
			cleanup_links(arch);
			return false;
			}
	return true;
	}

static bool del_link(const struct link *l)
	{
	bool success = false;
	char *target;
	ssize_t len;
	ino_t inode;


	if(!get_inode(&inode, l->file.p))
		/* link deleted? */
		return errno == ENOENT || err(MSG_STAT, true, l->file.p);
	if(l->inode != inode)  /* link replaced? */
		return true;

	target = alloc(plen(l->target) + 1);
	len = readlink(l->file.p, target, plen(l->target));
	if(len < 0)
		err(MSG_READLINK, true, l->file.p);
	else if(strncmp(target, l->target.p, len) == 0 &&
	        unlink(l->file.p) != 0 &&
	        errno != ENOENT)
		err(MSG_DEL, true, l->file.p);
	else
		success = true;
	free(target);
	return success;
	}

bool del_links(const struct archive *arch)
	{
	bool success = true;
	struct link *l;

	if(arch->links.file.p)
		success = del_link(&arch->links);
	for(l = arch->links.next; l; l = l->next)
		success = del_link(l) && success;
	return success;
	}
